import uuid
from typing import List

import msgspec
from litestar import Controller, get

from ..repositories import (
    UserRepository,
    provide_user_repository,
)


class PermissionRole(msgspec.Struct):
    id: uuid.UUID
    name: str


class Permission(msgspec.Struct):
    id: uuid.UUID
    name: str
    app: str


class Role(msgspec.Struct):
    id: uuid.UUID
    name: str
    permissions: List[Permission]


class User(msgspec.Struct):
    id: uuid.UUID
    email: str
    roles: List[Role]


class UserController(Controller):
    dependencies = {"user_repository": provide_user_repository}

    path = "/users"

    @get(path="/", exclude_from_auth=True)
    async def get_list(self, user_repository: UserRepository) -> List[User]:
        users = await user_repository.list()

        return [
            User(
                id=user.id,
                email=user.email,
                roles=[
                    Role(
                        id=role.id,
                        name=role.name,
                        permissions=[
                            Permission(
                                id=permission.id,
                                name=permission.name,
                                app=permission.app,
                            )
                            for permission in role.permissions
                        ],
                    )
                    for role in user.roles
                ],
            )
            for user in users
        ]



from advanced_alchemy.base import UUIDAuditBase
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .auth_association_models import role_permission_association, user_role_association


class RoleModel(UUIDAuditBase):
    __tablename__ = "roles"

    name: Mapped[str] = mapped_column(unique=True, nullable=False)
    users = relationship(
        "UserModel",
        secondary=user_role_association,
        back_populates="roles",
        lazy="selectin",
    )
    permissions = relationship(
        "PermissionModel",
        secondary=role_permission_association,
        back_populates="roles",
        lazy="selectin",
    )



from jwt.exceptions import (
    DecodeError,
    ExpiredSignatureError,
    InvalidAlgorithmError,
    InvalidSignatureError,
)
from litestar.connection import ASGIConnection
from litestar.exceptions import NotAuthorizedException
from litestar.middleware import AbstractAuthenticationMiddleware, AuthenticationResult
from litestar.middleware.base import DefineMiddleware

from .auth_core import decode_jwt_token

API_KEY_HEADER = "Authorization"
TOKEN_PREFIX = "Bearer "



class JWTAuthenticationMiddleware(AbstractAuthenticationMiddleware):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def authenticate_request(
        self,
        connection: ASGIConnection,
    ) -> AuthenticationResult:
        token = connection.headers.get(API_KEY_HEADER)

        if not token:
            raise NotAuthorizedException()

        token = token.replace(
            TOKEN_PREFIX,
            "",
        )
        try:
            token = decode_jwt_token(token)
        except (
            InvalidSignatureError,
            ExpiredSignatureError,
            InvalidAlgorithmError,
            DecodeError,
        ):
            raise NotAuthorizedException()

        user = token.payload
        if not user:
            raise NotAuthorizedException()
        return AuthenticationResult(
            user=user,
            auth=token,
        )


auth_middleware = DefineMiddleware(
    JWTAuthenticationMiddleware,
    exclude="/docs",
)



from .auth_association_models import role_permission_association, user_role_association
from .permission_model import PermissionModel
from .role_model import RoleModel
from .user_model import UserModel



import uuid
from typing import List

import msgspec
from litestar import Controller, get

from ..repositories import (
    RoleRepository,
    provide_role_repository,
)


class Permission(msgspec.Struct):
    id: uuid.UUID
    app: str
    name: str


class Role(msgspec.Struct):
    id: uuid.UUID
    name: str
    permissions: List[Permission]


class RoleController(Controller):
    dependencies = {"role_repository": provide_role_repository}

    path = "/roles"

    @get(path="/")
    async def get_list(self, role_repository: RoleRepository) -> List[Role]:
        roles = await role_repository.list()

        return [
            Role(
                id=role.id,
                name=role.name,
                permissions=[
                    Permission(
                        id=permission.id, name=permission.name, app=permission.app
                    )
                    for permission in role.permissions
                ],
            )
            for role in roles
        ]



from .permission_repository import PermissionRepository, provide_permission_repository
from .role_repository import RoleRepository, provide_role_repository
from .user_repository import UserRepository, provide_user_repository



import os

import click
import yaml

from ..core.cli import coro
from ..core.database import session_maker
from .auth_core import hash_password
from .models import PermissionModel, RoleModel, UserModel
from .repositories.permission_repository import provide_permission_repository
from .repositories.role_repository import provide_role_repository
from .repositories.user_repository import provide_user_repository


@click.group()
def auth_cli(): ...

@auth_cli.command()
@coro
async def load_auth():
    async with session_maker() as session:
        permission_repository = await provide_permission_repository(session)
        role_repository = await provide_role_repository(session)
        user_repository = await provide_user_repository(session)

        await permission_repository.delete_where(auto_commit=True)
        await role_repository.delete_where(auto_commit=True)
        await user_repository.delete_where(auto_commit=True)

        with open(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "./fixtures/auth.yaml"), "r"
        ) as stream:
            data_loaded = yaml.safe_load(stream)
            permissions = await permission_repository.add_many(
                [
                    PermissionModel(app=app, name=name)
                    for app, names in data_loaded.get("permissions").items()
                    for name in names
                ],
                auto_commit=True,
            )

            roles = []
            for name, permission_values in data_loaded.get("roles").items():
                role = RoleModel(name=name)
                role_permission_filter = {
                    app: {name for name in names}
                    for app, names in permission_values.items()
                }
                for permission in permissions:
                    if permission.name in role_permission_filter.get(permission.app):
                        role.permissions.append(permission)
                roles.append(role)

            await role_repository.add_many(roles, auto_commit=True)

            users = []

            for user_data in data_loaded.get("users"):
                email = user_data.get("email")
                users_roles = {role for role in user_data.get("roles")}

                user = UserModel(email=email, password=hash_password("password"))
                for role in roles:
                    if role.name in users_roles:
                        user.roles.append(role)
                users.append(user)

            await user_repository.add_many(users, auto_commit=True)



from litestar.openapi import OpenAPIConfig
from litestar.openapi.plugins import ScalarRenderPlugin
from litestar.openapi.spec import Components, SecurityScheme

openapi_config = OpenAPIConfig(
        components=[
            Components(
                security_schemes={
                    "JWT": SecurityScheme(
                        type="http",
                        scheme="Bearer",
                        name="Authorization",
                        security_scheme_in="cookie",
                        bearer_format="JWT",
                        description="Authorization",
                    )
                }
            )
        ],
        title="Service template",
        version="1.0.0",
        path="/docs",
        root_schema_site="elements",
        render_plugins=[ScalarRenderPlugin()],
    )



"""
Configure litestar app with lab module:

from litestar import Litestar

from lab.auth.api import auth_router
from lab.auth.middleware import auth_middleware
from lab.core.database import alchemy
from lab.core.openapi import openapi_config

app = Litestar(
    route_handlers=[auth_router],
    plugins=[alchemy],
    middleware=[auth_middleware],
    openapi_config=openapi_config,
)
"""


from datetime import datetime, timedelta
from typing import Any

import bcrypt
import jwt
from pydantic import BaseModel

from ..settings import settings

DEFAULT_TIME_DELTA = timedelta(days=30)
ALGORITHM = "HS256"


class Token(BaseModel):
    exp: datetime
    sub: str
    payload: Any


def decode_jwt_token(
    token: str,
) -> Token:
    return Token(
        **jwt.decode(
            token,
            settings.secret_key,
            algorithms=[ALGORITHM],
        )
    )


def encode_jwt_token(
    payload,
    expiration: timedelta = DEFAULT_TIME_DELTA,
) -> str:
    token = Token(
        exp=datetime.now() + expiration,
        sub="auth",
        payload=payload,
    )
    return jwt.encode(
        {
            "exp": token.exp,
            "sub": token.sub,
            "payload": token.payload,
        },
        settings.secret_key,
        algorithm=ALGORITHM,
    )


def hash_password(
    password: str,
) -> str:
    return bcrypt.hashpw(
        password.encode("utf-8"),
        bcrypt.gensalt(),
    ).decode("utf-8")


def check_password(
    password: str,
    hashed_password: str,
) -> bool:
    try:
        if bcrypt.checkpw(
            password.encode("utf-8"),
            hashed_password.encode("utf-8"),
        ):
            return True
    except ValueError:
        return False

    return False



import asyncio
from logging.config import fileConfig

from alembic import context
from litestar.contrib.sqlalchemy.base import UUIDBase
from sqlalchemy import engine_from_config, pool
from sqlalchemy.ext.asyncio import AsyncEngine

from lab.settings import settings


config = context.config

if config.config_file_name is not None:
    fileConfig(config.config_file_name)

target_metadata = UUIDBase.metadata
config.set_main_option(
    "sqlalchemy.url",
    settings.database_uri,
)


def run_migrations_offline() -> None:
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection):
    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        compare_type=True,
    )
    with context.begin_transaction():
        context.run_migrations()


async def run_async_migrations(connectable):
    async with connectable.connect() as connection:
        await connection.run_sync(do_run_migrations)
    await connectable.dispose()


def run_migrations_online():
    """
    Use in migrations/env.py alembic

    from app.main import app  # noqa
    from lab.core.migrations import run_migrations_online

    run_migrations_online()
    """
    connectable = context.config.attributes.get("connection", None)
    if connectable is None:
        connectable = AsyncEngine(
            engine_from_config(
                context.config.get_section(context.config.config_ini_section),
                prefix="sqlalchemy.",
                poolclass=pool.NullPool,
                future=True,
            )
        )

    if isinstance(connectable, AsyncEngine):
        asyncio.run(run_async_migrations(connectable))
    else:
        do_run_migrations(connectable)




from advanced_alchemy.base import UUIDAuditBase
from sqlalchemy import Column, ForeignKey, Table

user_role_association = Table(
    "users_roles",
    UUIDAuditBase.metadata,
    Column("user_id", ForeignKey("users.id", ondelete="CASCADE")),
    Column("role_id", ForeignKey("roles.id", ondelete="CASCADE")),
)

role_permission_association = Table(
    "roles_permissions",
    UUIDAuditBase.metadata,
    Column("role_id", ForeignKey("roles.id", ondelete="CASCADE")),
    Column("permission_id", ForeignKey("permissions.id", ondelete="CASCADE")),
)
